from typing import Optional, Any
from . import auth as syn


class DownloadStation:

    def __init__(self,
                 ip_address: str,
                 port: str,
                 username: str,
                 password: str,
                 secure: bool = False,
                 cert_verify: bool = False,
                 dsm_version: int = 7,
                 debug: bool = True,
                 otp_code: Optional[str] = None,
                 interactive_output: bool = True
                 ) -> None:

        self.session: syn.Authentication = syn.Authentication(ip_address, port, username, password, secure, cert_verify,
                                                              dsm_version, debug, otp_code)
        self._bt_search_id: str = ''
        self._bt_search_id_list: list[str] = []
        self.session.login('DownloadStation')
        self.session.get_api_list('DownloadStation')

        self.request_data: Any = self.session.request_data
        self.download_list: Any = self.session.app_api_list
        self._sid: str = self.session.sid
        self.base_url: str = self.session.base_url

        self.interactive_output: bool = interactive_output
        return

    def logout(self) -> None:
        self.session.logout('DownloadStation')
        return

    def get_info(self) -> dict[str, object] | str:
        api_name = 'SYNO.DownloadStation.Info'
        info = self.download_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'getinfo'}

        return self.request_data(api_name, api_path, req_param)

    def get_config(self) -> dict[str, object] | str:
        api_name = 'SYNO.DownloadStation.Info'
        info = self.download_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'getconfig'}

        return self.request_data(api_name, api_path, req_param)

    def set_server_config(self,
                          bt_max_download: Optional[int] = None,
                          bt_max_upload: Optional[int] = None,
                          emule_max_download: Optional[int] = None,
                          emule_max_upload: Optional[int] = None,
                          nzb_max_download: Optional[int] = None,
                          http_max_download: Optional[int] = None,
                          ftp_max_download: Optional[int] = None,
                          emule_enabled: Optional[bool] = None,
                          unzip_service_enabled: Optional[bool] = None,
                          default_destination: Optional[str] = None,
                          emule_default_destination: Optional[str] = None
                          ) -> dict[str, object] | str:

        api_name = 'SYNO.DownloadStation.Info'
        info = self.download_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'setserverconfig'}

        for key, val in locals().items():
            if key not in ['self', 'api_name', 'info', 'api_path', 'req_param']:
                if val is not None:
                    req_param[str(key)] = val

        return self.request_data(api_name, api_path, req_param)

    def schedule_info(self) -> dict[str, object] | str:
        api_name = 'SYNO.DownloadStation.Schedule'
        info = self.download_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'getconfig'}

        return self.request_data(api_name, api_path, req_param)

    def schedule_set_config(self, enabled: bool = False, emule_enabled: bool = False) -> dict[str, object] | str:
        api_name = 'SYNO.DownloadStation.Schedule'
        info = self.download_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'setconfig', 'enabled': str(enabled).lower(),
                     'emule_enabled': str(emule_enabled).lower()}

        if type(enabled) is not bool or type(emule_enabled) is not bool:
            return 'Please set enabled to True or False'

        return self.request_data(api_name, api_path, req_param)

    def tasks_list(self,
                   additional_param: Optional[str | list[str]] = None,
                   offset: int = 0,
                   limit: int = -1
                   ) -> dict[str, object] | str:
        api_name = 'SYNO.DownloadStation.Task'
        info = self.download_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'list', 'additional': additional_param, 'limit': limit,
                     'offset': offset}

        if additional_param is None:
            additional_param = ['detail', 'transfer', 'file', 'tracker', 'peer']

        if type(additional_param) is list:
            req_param['additional'] = ",".join(additional_param)

        return self.request_data(api_name, api_path, req_param)

    def tasks_info(self, task_id, additional_param: Optional[str | list[str]] = None) -> dict[str, object] | str:
        api_name = 'SYNO.DownloadStation.Task'
        info = self.download_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'getinfo', 'id': task_id, 'additional': additional_param}

        if additional_param is None:
            additional_param = ['detail', 'transfer', 'file', 'tracker', 'peer']

        if type(additional_param) is list:
            req_param['additional'] = ",".join(additional_param)

        if type(task_id) is list:
            req_param['id'] = ",".join(task_id)

        return self.request_data(api_name, api_path, req_param)

    def create_task(self, uri, additional_param: Optional[dict[str, object]] = None) -> dict[str, object] | str:
        api_name = 'SYNO.DownloadStation.Task'
        info = self.download_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'create', 'uri': uri}

        if type(additional_param) is dict:
            for key in additional_param.keys():
                req_param[key] = additional_param[key]

        return self.request_data(api_name, api_path, req_param)

    def delete_task(self, task_id: str, force: bool = False) -> dict[str, object] | str:
        api_name = 'SYNO.DownloadStation.Task'
        info = self.download_list[api_name]
        api_path = info['path']
        param = {'version': info['maxVersion'], 'method': 'delete', 'id': task_id,
                 'force_complete': str(force).lower()}

        if type(task_id) is list:
            param['id'] = ",".join(task_id)

        return self.request_data(api_name, api_path, param)

    def pause_task(self, task_id: str) -> dict[str, object] | str:
        api_name = 'SYNO.DownloadStation.Task'
        info = self.download_list[api_name]
        api_path = info['path']
        param = {'version': info['maxVersion'], 'method': 'pause', 'id': task_id}

        if type(task_id) is list:
            param['id'] = ",".join(task_id)

        return self.request_data(api_name, api_path, param)

    def resume_task(self, task_id: str) -> dict[str, object] | str:
        api_name = 'SYNO.DownloadStation.Task'
        info = self.download_list[api_name]
        api_path = info['path']
        param = {'version': info['maxVersion'], 'method': 'resume', 'id': task_id}

        if type(task_id) is list:
            param['id'] = ",".join(task_id)

        return self.request_data(api_name, api_path, param)

    def edit_task(self, task_id: str, destination: str = 'sharedfolder') -> dict[str, object] | str:
        api_name = 'SYNO.DownloadStation.Task'
        info = self.download_list[api_name]
        api_path = info['path']
        param = {'version': info['maxVersion'], 'method': 'edit', 'id': task_id, 'destination': destination}

        if type(task_id) is list:
            param['id'] = ",".join(task_id)

        return self.request_data(api_name, api_path, param)

    def get_statistic_info(self) -> dict[str, object] | str:
        api_name = 'SYNO.DownloadStation.Statistic'
        info = self.download_list[api_name]
        api_path = info['path']
        param = {'version': info['maxVersion'], 'method': 'getinfo'}

        return self.request_data(api_name, api_path, param)

    def get_rss_info_list(self, offset: Optional[int] = None, limit: Optional[int] = None) -> dict[str, object] | str:
        api_name = 'SYNO.DownloadStation.RSS.Site'
        info = self.download_list[api_name]
        api_path = info['path']
        param = {'version': info['maxVersion'], 'method': 'list'}

        if offset is not None:
            param['offset'] = offset
        if limit is not None:
            param['limit'] = limit

        return self.request_data(api_name, api_path, param)

    def refresh_rss_site(self, rss_id: Optional[str] = None) -> dict[str, object] | str:
        api_name = 'SYNO.DownloadStation.RSS.Site'
        info = self.download_list[api_name]
        api_path = info['path']
        param = {'version': info['maxVersion'], 'method': 'refresh', 'id': rss_id}

        if rss_id is None:
            return 'Enter a valid ID check if you have any with get_rss_list()'
        elif type(rss_id) is list:
            rss_id = ','.join(rss_id)
            param['id'] = rss_id

        return self.request_data(api_name, api_path, param)

    def rss_feed_list(self,
                      rss_id: Optional[str] = None,
                      offset: Optional[int] = None,
                      limit: Optional[int] = None
                      ) -> dict[str, object] | str:
        api_name = 'SYNO.DownloadStation.RSS.Feed'
        info = self.download_list[api_name]
        api_path = info['path']
        param = {'version': info['maxVersion'], 'method': 'list', 'id': rss_id}

        if rss_id is None:
            return 'Enter a valid ID check if you have any with get_rss_list()'
        elif type(rss_id) is list:
            rss_id = ','.join(rss_id)
            param['id'] = rss_id

        if offset is not None:
            param['offset'] = offset
        if limit is not None:
            param['limit'] = limit

        return self.request_data(api_name, api_path, param)

    def start_bt_search(self, keyword: Optional[str] = None, module: str = 'all') -> dict[str, object] | str:
        api_name = 'SYNO.DownloadStation.BTSearch'
        info = self.download_list[api_name]
        api_path = info['path']
        param = {'version': info['maxVersion'], 'method': 'start'}

        if keyword is None:
            return 'Did you enter a keyword to search?'
        else:
            param['keyword'] = keyword

        param['module'] = module

        self._bt_search_id = self.request_data(api_name, api_path, param)['data']['taskid']

        self._bt_search_id_list.append(self._bt_search_id)

        message = ('You can now check the status of request with '
                   'get_bt_search_results(), your id is: '
                   + self._bt_search_id)
        if self.interactive_output:
            output = message
        else:
            output = {"message": message, "taskid": self._bt_search_id}

        return output

    def get_bt_search_results(self,
                              taskid: Optional[str] = None,
                              offset: Optional[int] = None,
                              limit: Optional[int] = None,
                              sort_by: Optional[str] = None,
                              sort_direction: Optional[str] = None,
                              filter_category: Optional[str] = None,
                              filter_title: Optional[str] = None
                              ) -> dict[str, object] | str:
        api_name = 'SYNO.DownloadStation.BTSearch'
        info = self.download_list[api_name]
        api_path = info['path']
        param = {'version': info['maxVersion'], 'method': 'list', 'taskid': taskid}

        for key, val in locals().items():
            if key not in ['self', 'api_name', 'info', 'api_path', 'param', 'taskid']:
                if val is not None:
                    param[str(key)] = val

        if taskid is None:
            return 'Enter a valid taskid, you can choose one of ' + str(self._bt_search_id_list)
        elif type(taskid) is list:
            param['taskid'] = ','.join(taskid)

        return self.request_data(api_name, api_path, param)

    def get_bt_search_category(self) -> dict[str, object] | str:
        api_name = 'SYNO.DownloadStation.BTSearch'
        info = self.download_list[api_name]
        api_path = info['path']
        param = {'version': info['maxVersion'], 'method': 'get'}

        return self.request_data(api_name, api_path, param)

    def clean_bt_search(self, taskid: Optional[str | list[str]] = None) -> dict[str, object] | str:
        api_name = 'SYNO.DownloadStation.BTSearch'
        info = self.download_list[api_name]
        api_path = info['path']
        param = {'version': info['maxVersion'], 'method': 'clean', 'taskid': taskid}

        if taskid is None:
            return 'Enter a valid taskid, you can choose one of ' + str(self._bt_search_id_list)
        elif type(taskid) is list:
            param['taskid'] = ','.join(taskid)
            for item in taskid:
                self._bt_search_id_list.remove(item)
        else:
            self._bt_search_id_list.remove(taskid)

        return self.request_data(api_name, api_path, param)

    def get_bt_module(self) -> dict[str, object] | str:
        api_name = 'SYNO.DownloadStation.BTSearch'
        info = self.download_list[api_name]
        api_path = info['path']
        param = {'version': info['maxVersion'], 'method': 'getModule'}

        return self.request_data(api_name, api_path, param)



from setuptools import setup, find_packages

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='synology-api',
    version='0.5.2',
    packages=find_packages(exclude=['tests*']),
    license='MIT',
    description='Python Synology API Wrapper',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=['requests', 'urllib3', 'setuptools'],
    url='https://github.com/N4S4/synology-api',
    author='Renato Visaggio',
    author_email='synology.python.api@gmail.com'
)



from typing import Optional
from . import base_api_core


class USBCopy(base_api_core.Core):
    def __init__(self,
                 ip_address: str,
                 port: str,
                 username: str,
                 password: str,
                 secure: bool = False,
                 cert_verify: bool = False,
                 dsm_version: int = 7,
                 debug: bool = True,
                 otp_code: Optional[str] = None
                 ) -> None:
        super(USBCopy, self).__init__(ip_address, port, username, password, secure, cert_verify, dsm_version, debug,
                                      otp_code)
        return

    def usb_copy_info(self, id: int = 1) -> dict[str, object] | str:
        api_name = 'SYNO.USBCopy'
        info = self.gen_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'get', 'id': id}

        return self.request_data(api_name, api_path, req_param)

    def toggle_usb_copy(self, enable: bool = True, id: int = 1) -> dict[str, object] | str:
        api_name = 'SYNO.USBCopy'
        info = self.gen_list[api_name]
        api_path = info['path']

        if enable:
            enable = 'enable'
        elif not enable:
            enable = 'disable'
        else:
            return 'enable must be True or False'

        req_param = {'version': info['maxVersion'], 'method': enable, 'id': id}

        return self.request_data(api_name, api_path, req_param)

    def logs(self, offset: int = 0, limit: int = 200) -> dict[str, object] | str:
        api_name = 'SYNO.USBCopy'
        info = self.gen_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'get_log_list', 'offset': offset, 'limit': limit,
                     'log_filter': '{"log_desc_id_list":[0,1,2,3,10,11,100,101,102,103,104,105,1000]}'}

        return self.request_data(api_name, api_path, req_param)

    def global_settings(self) -> dict[str, object] | str:
        api_name = 'SYNO.USBCopy'
        info = self.gen_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'get_global_setting'}

        return self.request_data(api_name, api_path, req_param)



from typing import Optional, Any
import os
import time
from datetime import datetime

import requests
import sys
from urllib import parse

from . import auth as syn


class FileStation:

    def __init__(self,
                 ip_address: str,
                 port: str,
                 username: str,
                 password: str,
                 secure: bool = False,
                 cert_verify: bool = False,
                 dsm_version: int = 7,
                 debug: bool = True,
                 otp_code: Optional[str] = None,
                 interactive_output: bool = True
                 ) -> None:

        self.session: syn.Authentication = syn.Authentication(ip_address, port, username, password, secure, cert_verify,
                                                              dsm_version, debug, otp_code)

        self._dir_taskid: str = ''
        self._dir_taskid_list: list[str] = []
        self._md5_calc_taskid: str = ''
        self._md5_calc_taskid_list: list[str] = []
        self._search_taskid: str = ''
        self._search_taskid_list: list[str] = []
        self._copy_move_taskid: str = ''
        self._copy_move_taskid_list: list[str] = []
        self._delete_taskid: str = ''
        self._delete_taskid_list: list[str] = []
        self._extract_taskid: str = ''
        self._extract_taskid_list: list[str] = []
        self._compress_taskid: str = ''
        self._compress_taskid_list: list[str] = []
        self.request_data = self.session.request_data

        self.session.login('FileStation')
        self.session.get_api_list('FileStation')

        self.file_station_list: Any = self.session.app_api_list
        self._sid: str = self.session.sid
        self.base_url: str = self.session.base_url

        self.interactive_output: bool = interactive_output

    def logout(self) -> None:
        self.session.logout('FileStation')
        return

    def get_info(self) -> dict[str, object] | str:
        api_name = 'SYNO.FileStation.Info'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'get', '_sid': self._sid}

        return self.request_data(api_name, api_path, req_param)

    def get_list_share(self,
                       additional: Optional[str | list[str]] = None,
                       offset: Optional[int] = None,
                       limit: Optional[int] = None,
                       sort_by: Optional[str] = None,
                       sort_direction: Optional[str] = None,
                       onlywritable: bool = False
                       ) -> dict[str, object] | str:

        api_name = 'SYNO.FileStation.List'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'list_share'}

        for key, val in locals().items():
            if key not in ['self', 'api_name', 'info', 'api_path', 'req_param', 'additional']:
                if val is not None:
                    req_param[str(key)] = val

        if additional is None:
            additional = ['real_path', 'size', 'owner', 'time']

        if type(additional) is list:
            additional = ','.join(additional)

        req_param['additional'] = additional

        return self.request_data(api_name, api_path, req_param)

    def get_file_list(self,
                      folder_path: Optional[str] = None,
                      offset: Optional[int] = None,
                      limit: Optional[int] = None,
                      sort_by: Optional[str] = None,
                      sort_direction: Optional[str] = None,
                      pattern: Optional[str] = None,
                      filetype: Optional[str] = None,
                      goto_path: Optional[str] = None,
                      additional: Optional[str | list[str]] = None) -> dict[str, object] | str:

        api_name = 'SYNO.FileStation.List'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'list'}

        for key, val in locals().items():
            if key not in ['self', 'api_name', 'info', 'api_path', 'req_param', 'additional']:
                if val is not None:
                    req_param[str(key)] = val

        if folder_path is None:
            return 'Enter a valid folder_path'

        if filetype is not None:
            req_param['filetype'] = str(req_param['filetype']).lower()

        if additional is None:
            additional = ['real_path', 'size', 'owner', 'time']

        if type(additional) is list:
            additional = ','.join(additional)

        req_param['additional'] = additional

        return self.request_data(api_name, api_path, req_param)

    def get_file_info(self,
                      path: Optional[str] = None,
                      additional: Optional[str | list[str]] = None
                      ) -> dict[str, object] | str:
        api_name = 'SYNO.FileStation.List'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'getinfo'}

        if type(path) is list:
            new_path = []
            [new_path.append('"' + x + '"') for x in path]
            path = new_path
            path = '[' + ','.join(path) + ']'
            req_param['path'] = path
        elif path is not None:
            req_param['path'] = path

        if additional is None:
            additional = ['real_path', 'size', 'owner', 'time']

        if type(additional) is list:
            additional = str(additional).replace("'", '"')

        req_param['additional'] = additional

        return self.request_data(api_name, api_path, req_param)

    # TODO  all working if specify extension check if correct [pattern, extension]
    #  it works if you put extension='...'

    def search_start(self,
                     folder_path: Optional[str] = None,
                     recursive: Optional[bool] = None,
                     pattern: Optional[str] = None,
                     extension: Optional[str] = None,
                     filetype: Optional[str] = None,
                     size_from: Optional[int] = None,
                     size_to: Optional[int] = None,
                     mtime_from: Optional[str | int] = None,
                     mtime_to: Optional[str | int] = None,
                     crtime_from: Optional[str | int] = None,
                     crtime_to: Optional[str | int] = None,
                     atime_from: Optional[str | int] = None,
                     atime_to: Optional[str | int] = None,
                     owner: Optional[str] = None,
                     group: Optional[str] = None
                     ) -> dict[str, object] | str:

        api_name = 'SYNO.FileStation.Search'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'start', 'folder_path': ''}

        for key, val in locals().items():
            if key not in ['self', 'api_name', 'info', 'api_path', 'req_param'] and 'time' not in key:
                if val is not None:
                    req_param[str(key)] = val
            if 'time' in key:
                if val is not None:
                    try:
                        date = time.strptime(val, "%Y-%m-%d %H:%M:%S")
                        timestamp = time.mktime(date)
                        req_param[key] = '"' + str(timestamp) + '"'
                    except ValueError:
                        try:
                            datetime.fromtimestamp(int(val)).strftime('%Y-%m-%d %H:%M:%S')
                            req_param[key] = '"' + val + '"'
                        except ValueError:
                            return 'Enter the correct Date Time format "YYY-MM-DD HH:MM:SS" or Unix timestamp'

        if folder_path is None:
            return 'Enter a valid folder_path'
        else:
            req_param['folder_path'] = '"' + folder_path + '"'

        if filetype is not None:
            req_param['filetype'] = '"' + filetype + '"'

        response = self.request_data(api_name, api_path, req_param)

        taskid = response['data']['taskid']
        self._search_taskid = '"{}"'.format(taskid)
        self._search_taskid_list.append('"' + response['data']['taskid'] + '"')

        message = ('You can now check the status of request with '
                   'get_search_list() , your id is: ' + self._search_taskid)
        if self.interactive_output:
            output = message
        else:
            output = {"message": message, "taskid": taskid}

        return output

    def get_search_list(self,
                        task_id: str,
                        filetype: Optional[str] = None,
                        limit: Optional[int] = None,
                        sort_by: Optional[str] = None,
                        sort_direction: Optional[str] = None,
                        offset: Optional[int] = None,
                        additional: Optional[str | list[str]] = None
                        ) -> dict[str, object] | str:
        api_name = 'SYNO.FileStation.Search'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'list', 'taskid': ''}

        if task_id is None:
            return 'Enter a correct taskid, choose one of the following: ' + str(self._search_taskid_list)
        else:
            req_param['taskid'] = task_id

        for key, val in locals().items():
            if key not in ['self', 'api_name', 'info', 'api_path', 'req_param', 'additional', 'task_id']:
                if val is not None:
                    req_param[str(key)] = val

        if filetype is not None:
            req_param['filetype'] = str(filetype).lower()

        if additional is None:
            additional = ['size', 'owner', 'time']

        if type(additional) is list:
            additional = '","'.join(additional)

        req_param['additional'] = '["' + additional + '"]'

        return self.request_data(api_name, api_path, req_param)

    def stop_search_task(self, taskid: str) -> dict[str, object] | str:
        api_name = 'SYNO.FileStation.Search'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'stop', 'taskid': self._search_taskid}

        if taskid is None:
            return 'Enter a valid taskid, choose between ' + str(self._search_taskid_list)

        self._search_taskid_list.remove(taskid)

        return self.request_data(api_name, api_path, req_param)

    def stop_all_search_task(self) -> str:
        api_name = 'SYNO.FileStation.Search'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'stop', 'taskid': ''}

        assert len(self._search_taskid_list), 'Task list is empty' + str(self._search_taskid_list)

        for task_id in self._search_taskid_list:
            req_param['taskid'] = task_id
            self.request_data(api_name, api_path, req_param)

        self._search_taskid_list = []

        return 'All task are stopped'

    def get_mount_point_list(self,
                             mount_type: Optional[str] = None,
                             offset: Optional[int] = None,
                             limit: Optional[int] = None,
                             sort_by: Optional[str] = None,
                             sort_direction: Optional[str] = None,
                             additional: Optional[str | list[str]] = None
                             ) -> dict[str, object] | str:

        api_name = 'SYNO.FileStation.VirtualFolder'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'list'}

        if mount_type is not None:
            req_param['type'] = mount_type

        for key, val in locals().items():
            if key not in ['self', 'api_name', 'info', 'api_path', 'req_param', 'additional', 'mount_type']:
                if val is not None:
                    req_param[str(key)] = val

        if additional is None:
            additional = ['real_path', 'size', 'owner', 'time']

        if type(additional) is list:
            additional = ','.join(additional)

        req_param['additional'] = additional

        return self.request_data(api_name, api_path, req_param)

    def get_favorite_list(self,
                          offset: Optional[int] = None,
                          limit: Optional[int] = None,
                          sort_by: Optional[str] = None,
                          status_filter: Optional[str] = None,
                          additional: Optional[str | list[str]] = None
                          ) -> dict[str, object] | str:

        api_name = 'SYNO.FileStation.Favorite'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'list'}

        for key, val in locals().items():
            if key not in ['self', 'api_name', 'info', 'api_path', 'req_param', 'additional']:
                if val is not None:
                    req_param[str(key)] = val

        if additional is None:
            additional = ['real_path', 'size', 'owner', 'time']

        if type(additional) is list:
            additional = ','.join(additional)

        req_param['additional'] = additional

        return self.request_data(api_name, api_path, req_param)

    def add_a_favorite(self,
                       path: str,
                       name: Optional[str] = None,
                       index: Optional[int] = None
                       ) -> dict[str, object] | str:
        api_name = 'SYNO.FileStation.Favorite'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'add'}

        if path is None:
            return 'Enter a valid path'

        for key, val in locals().items():
            if key not in ['self', 'api_name', 'info', 'api_path', 'req_param']:
                if val is not None:
                    req_param[str(key)] = val

        return self.request_data(api_name, api_path, req_param)

    def delete_a_favorite(self, path: Optional[str] = None) -> dict[str, object] | str:
        api_name = 'SYNO.FileStation.Favorite'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'delete'}

        for key, val in locals().items():
            if key not in ['self', 'api_name', 'info', 'api_path', 'req_param']:
                if val is not None:
                    req_param[str(key)] = val

        return self.request_data(api_name, api_path, req_param)

    def clear_broken_favorite(self) -> dict[str, object] | str:
        api_name = 'SYNO.FileStation.Favorite'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'clear_broken'}

        return self.request_data(api_name, api_path, req_param)

    def edit_favorite_name(self, path: str, new_name: str) -> dict[str, object] | str:
        api_name = 'SYNO.FileStation.Favorite'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'edit'}

        if path is None:
            return 'Enter a valid path'
        else:
            req_param['path'] = path

        if new_name is None:
            return 'Enter a valid new_name'
        else:
            req_param['new_name'] = new_name

        return self.request_data(api_name, api_path, req_param)

    def replace_all_favorite(self, path: str | list[str], name: str | list[str]):
        api_name = 'SYNO.FileStation.Favorite'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'edit'}

        if type(path) is list:
            path = ','.join(path)
            req_param['path'] = path
        elif path is not None:
            req_param['path'] = path
        else:
            return 'Enter a valid path'

        if type(name) is list:
            name = ','.join(name)
            req_param['name'] = name
        elif name is not None:
            req_param['name'] = name
        else:
            return 'Enter a valid name'

        return self.request_data(api_name, api_path, req_param)

    def start_dir_size_calc(self, path: str) -> dict[str, object] | str:
        api_name = 'SYNO.FileStation.DirSize'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'start'}

        if type(path) is list:
            new_path = []
            [new_path.append('"' + x + '"') for x in path]
            path = new_path
            path = '[' + ','.join(path) + ']'
            req_param['path'] = path
        elif path is not None:
            req_param['path'] = path
        else:
            return 'Enter a valid path'

        taskid = self.request_data(api_name, api_path, req_param)['data']['taskid']

        response_id = '"{}"'.format(taskid)
        self._dir_taskid = response_id
        self._dir_taskid_list.append(response_id)

        message = ('You can now check the status of request '
                   'with get_dir_status() , your id is: '
                   + response_id)
        if self.interactive_output:
            output = message
        else:
            output = {"message": message, "taskid": taskid}

        return output

    def stop_dir_size_calc(self, taskid: str) -> str:
        api_name = 'SYNO.FileStation.DirSize'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'stop', 'taskid': taskid}

        if taskid is None:
            return 'Enter a valid taskid choose between: ' + str(self._dir_taskid_list)
        else:
            req_param['taskid'] = '"' + taskid + '"'

        self.request_data(api_name, api_path, req_param)
        self._dir_taskid_list.remove('"' + taskid + '"')

        return 'The task has been stopped'

    def get_dir_status(self, taskid: Optional[str] = None) -> dict[str, object] | str:
        api_name = 'SYNO.FileStation.DirSize'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'status', 'taskid': taskid}

        if taskid is None and self._dir_taskid != '':
            return 'Choose a taskid from this list: ' + str(self._dir_taskid)

        return self.request_data(api_name, api_path, req_param)

    def start_md5_calc(self, file_path: str) -> str | dict[str, object]:
        api_name = 'SYNO.FileStation.MD5'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'start'}

        if file_path is None:
            return 'Enter a correct file_path'
        else:
            req_param['file_path'] = file_path

        self._md5_calc_taskid = self.request_data(api_name, api_path, req_param)['data']['taskid']
        self._md5_calc_taskid_list.append(self._md5_calc_taskid)

        message = ('You can now check the status of request with '
                   'get_md5_status() , your id is: ' + self._md5_calc_taskid)
        if self.interactive_output:
            output = message
        else:
            output = {"message": message, "taskid": self._md5_calc_taskid}

        return output

    def get_md5_status(self, taskid: Optional[str] = None) -> str | dict[str, object]:
        api_name = 'SYNO.FileStation.MD5'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'status'}

        if taskid is None and self._md5_calc_taskid != '':
            req_param['taskid'] = '"{taskid}"'.format(taskid=self._md5_calc_taskid)
        elif taskid is not None:
            req_param['taskid'] = '"{taskid}"'.format(taskid=taskid)
        else:
            return 'Did you run start_md5_calc() first? No task id found! ' + str(self._md5_calc_taskid)

        return self.request_data(api_name, api_path, req_param)

    def stop_md5_calc(self, taskid: str) -> str:
        api_name = 'SYNO.FileStation.DirSize'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'stop', 'taskid': taskid}

        if taskid is None:
            return 'Enter a valid taskid choose between: ' + str(self._md5_calc_taskid_list)
        else:
            req_param['taskid'] = '"' + taskid + '"'

        self.request_data(api_name, api_path, req_param)
        self._md5_calc_taskid_list.remove(taskid)

        return 'The task has been stopped'

    def check_permissions(self,
                          path: str,
                          filename: str,
                          overwrite: Optional[bool] = None,
                          create_only: Optional[bool] = None
                          ) -> dict[str, object] | str:
        api_name = 'SYNO.FileStation.CheckPermission'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'write'}

        for key, val in locals().items():
            if key not in ['self', 'api_name', 'info', 'api_path', 'req_param']:
                if val is not None:
                    req_param[str(key)] = val

        if path is None:
            return 'Enter a valid path'

        if filename is None:
            return 'Enter a valid name'

        return self.request_data(api_name, api_path, req_param)

    def upload_file(self,
                    dest_path: str,
                    file_path: str,
                    create_parents: bool = True,
                    overwrite: bool = True,
                    verify: bool = False
                    ) -> str | tuple[int, dict[str, object]]:
        api_name = 'SYNO.FileStation.Upload'
        info = self.file_station_list[api_name]
        api_path = info['path']
        filename = os.path.basename(file_path)

        session = requests.session()

        with open(file_path, 'rb') as payload:
            url = ('%s%s' % (self.base_url, api_path)) + '?api=%s&version=%s&method=upload&_sid=%s' % (
                api_name, info['minVersion'], self._sid)

            args = {
                'path': dest_path,
                'create_parents': str(create_parents).lower(),
                'overwrite': str(overwrite).lower(),
            }

            files = {'file': (filename, payload, 'application/octet-stream')}

            r = session.post(url, data=args, files=files, verify=verify)

            if r.status_code == 200 and r.json()['success']:
                return 'Upload Complete'
            else:
                return r.status_code, r.json()

    def get_shared_link_info(self, link_id: str) -> dict[str, object] | str:
        api_name = 'SYNO.FileStation.Sharing'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'getinfo'}

        if link_id is None:
            return 'Enter a valid id'
        else:
            req_param['id'] = link_id

        return self.request_data(api_name, api_path, req_param)

    def get_shared_link_list(self,
                             offset: Optional[int] = None,
                             limit: Optional[int] = None,
                             sort_by: Optional[str] = None,
                             sort_direction: Optional[str] = None,
                             force_clean: Optional[bool] = None
                             ) -> dict[str, object] | str:

        api_name = 'SYNO.FileStation.Sharing'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'list'}

        for key, val in locals().items():
            if key not in ['self', 'api_name', 'info', 'api_path', 'req_param']:
                if val is not None:
                    req_param[str(key)] = val

        return self.request_data(api_name, api_path, req_param)

    def create_sharing_link(self,
                            path: str,
                            password: Optional[str] = None,
                            date_expired: Optional[str | int] = None,
                            date_available: Optional[str | int] = None,
                            expire_times: int = 0
                            ) -> dict[str, object] | str:
        api_name = 'SYNO.FileStation.Sharing'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'create'}

        for key, val in locals().items():
            if key not in ['self', 'api_name', 'info', 'api_path', 'req_param']:
                if val is not None:
                    req_param[str(key)] = val

        if path is None:
            return 'Enter a valid path'

        return self.request_data(api_name, api_path, req_param)

    def delete_shared_link(self, link_id: str) -> dict[str, object] | str:
        api_name = 'SYNO.FileStation.Sharing'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'delete'}

        if link_id is None:
            return 'Enter a valid id'
        else:
            req_param['id'] = link_id

        return self.request_data(api_name, api_path, req_param)

    def clear_invalid_shared_link(self) -> dict[str, object] | str:
        api_name = 'SYNO.FileStation.Sharing'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'clear_invalid'}

        return self.request_data(api_name, api_path, req_param)

    def edit_shared_link(self,
                         link_id: str,
                         password: Optional[str] = None,
                         date_expired: Optional[str | int] = None,
                         date_available: Optional[str | int] = None,
                         expire_times: int = 0
                         ) -> dict[str, object] | str:
        api_name = 'SYNO.FileStation.Sharing'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'edit'}

        if link_id is None:
            return 'Enter a valid id'
        else:
            req_param['id'] = link_id

        for key, val in locals().items():
            if key not in ['self', 'api_name', 'info', 'api_path', 'req_param']:
                if val is not None:
                    req_param[str(key)] = val

        return self.request_data(api_name, api_path, req_param)

    def create_folder(self,
                      folder_path: str | list[str],
                      name: str | list[str],
                      force_parent: Optional[bool] = None,
                      additional: Optional[str | list[str]] = None
                      ) -> str | dict[str, object]:
        api_name = 'SYNO.FileStation.CreateFolder'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'create'}

        for key, val in locals().items():
            if key not in ['self', 'api_name', 'info', 'api_path', 'req_param', 'folder_path', 'additional', 'name']:
                if val is not None:
                    req_param[str(key)] = val

        if type(folder_path) is list:
            new_path = []
            [new_path.append('"' + x + '"') for x in folder_path]
            folder_path = new_path
            folder_path = '[' + ','.join(folder_path) + ']'
            req_param['folder_path'] = folder_path
        elif folder_path is not None:
            req_param['folder_path'] = folder_path
        else:
            return 'Enter a valid path'

        if type(name) is list:
            new_path = []
            [new_path.append('"' + x + '"') for x in name]
            name = new_path
            name = '[' + ','.join(name) + ']'
            req_param['name'] = name
        elif name is not None:
            req_param['name'] = '"' + name + '"'
        else:
            return 'Enter a valid path'

        if additional is None:
            additional = ['real_path', 'size', 'owner', 'time']

        if type(additional) is list:
            additional = ','.join(additional)

        req_param['additional'] = additional

        return self.request_data(api_name, api_path, req_param)

    def rename_folder(self,
                      path: str | list[str],
                      name: str | list[str],
                      additional: Optional[str | list[str]] = None,
                      search_taskid: Optional[str] = None
                      ) -> dict[str, object] | str:
        api_name = 'SYNO.FileStation.Rename'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'rename'}

        if type(path) is list:
            new_path = []
            [new_path.append('"' + x + '"') for x in path]
            path = new_path
            path = '[' + ','.join(path) + ']'
            req_param['path'] = path
        elif path is not None:
            req_param['path'] = path
        else:
            return 'Enter a valid path'

        if type(name) is list:
            new_path = []
            [new_path.append('"' + x + '"') for x in name]
            name = new_path
            name = '[' + ','.join(name) + ']'
            req_param['name'] = name
        elif name is not None:
            req_param['name'] = name
        else:
            return 'Enter a valid path'

        if additional is None:
            additional = ['real_path', 'size', 'owner', 'time']

        if type(additional) is list:
            additional = ','.join(additional)

        req_param['additional'] = additional

        if search_taskid is not None:
            req_param['search_taskid'] = search_taskid

        return self.request_data(api_name, api_path, req_param)

    def start_copy_move(self,
                        path: str | list[str],
                        dest_folder_path: str | list[str],
                        overwrite: Optional[bool] = None,
                        remove_src: Optional[bool] = None,
                        accurate_progress: Optional[bool] = None,
                        search_taskid: Optional[str] = None
                        ) -> str | dict[str, object]:
        api_name = 'SYNO.FileStation.CopyMove'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'start'}

        if type(path) is list:
            new_path = []
            [new_path.append('"' + x + '"') for x in path]
            path = new_path
            path = '[' + ','.join(path) + ']'
            req_param['path'] = path
        elif path is not None:
            req_param['path'] = path
        else:
            return 'Enter a valid path'

        if type(dest_folder_path) is list:
            new_path = []
            [new_path.append('"' + x + '"') for x in dest_folder_path]
            dest_folder_path = new_path
            dest_folder_path = '[' + ','.join(dest_folder_path) + ']'
            req_param['name'] = dest_folder_path
        elif dest_folder_path is not None:
            req_param['dest_folder_path'] = dest_folder_path
        else:
            return 'Enter a valid path'

        for key, val in locals().items():
            if key not in ['self', 'api_name', 'info', 'api_path', 'req_param', 'path', 'additional',
                           'dest_folder_path', 'new_path']:
                if val is not None:
                    req_param[str(key)] = val

        self._copy_move_taskid = self.request_data(api_name, api_path, req_param)['data']['taskid']
        self._dir_taskid_list.append(self._copy_move_taskid)

        message = ('You can now check the status of request with '
                   'get_copy_move_status() , your id is: '
                   + self._copy_move_taskid)
        if self.interactive_output:
            output = message
        else:
            output = {"message": message, "taskid": self._copy_move_taskid}

        return output

    def get_copy_move_status(self, taskid: str) -> dict[str, object] | str:
        api_name = 'SYNO.FileStation.CopyMove'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'status'}

        if taskid is None:
            return 'Enter a valid taskid choose between ' + str(self._copy_move_taskid_list)
        else:
            req_param['taskid'] = '"' + taskid + '"'

        return self.request_data(api_name, api_path, req_param)

    def stop_copy_move_task(self, taskid: str) -> dict[str, object] | str:
        api_name = 'SYNO.FileStation.CopyMove'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'stop'}

        if taskid is None:
            return 'Enter a valid taskid choose between ' + str(self._copy_move_taskid_list)
        else:
            req_param['taskid'] = taskid

        self._copy_move_taskid_list.remove(taskid)

        return self.request_data(api_name, api_path, req_param)

    def start_delete_task(self,
                          path: str | list[str],
                          accurate_progress: Optional[bool] = None,
                          recursive: Optional[bool] = None,
                          search_taskid: Optional[str] = None
                          ) -> dict[str, object] | str:
        api_name = 'SYNO.FileStation.Delete'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'start'}

        if type(path) is list:
            new_path = []
            [new_path.append('"' + x + '"') for x in path]
            path = new_path
            path = '[' + ','.join(path) + ']'
            req_param['path'] = path
        elif path is not None:
            req_param['path'] = path
        else:
            return 'Enter a valid path'

        for key, val in locals().items():
            if key not in ['self', 'api_name', 'info', 'api_path', 'req_param', 'path', 'new_path']:
                if val is not None:
                    req_param[str(key)] = val

        self._delete_taskid = self.request_data(api_name, api_path, req_param)['data']['taskid']
        self._delete_taskid_list.append(self._delete_taskid)

        message = ('You can now check the status of request with '
                   'get_delete_status() , task id is: '
                   + self._delete_taskid)
        if self.interactive_output:
            output = message
        else:
            output = {"message": message, "taskid": self._delete_taskid}

        return output

    def get_delete_status(self, taskid: str) -> dict[str, object] | str:
        api_name = 'SYNO.FileStation.Delete'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'status'}

        if taskid is None:
            return 'Enter a valid taskid, choose between ' + str(self._delete_taskid_list)
        else:
            req_param['taskid'] = taskid

        return self.request_data(api_name, api_path, req_param)

    def stop_delete_task(self, taskid: str) -> dict[str, object] | str:
        api_name = 'SYNO.FileStation.Delete'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'stop'}

        if taskid is None:
            return 'Enter a valid taskid, choose between ' + str(self._delete_taskid_list)
        else:
            req_param['taskid'] = taskid

        self._delete_taskid_list.remove('"' + taskid + '"')

        return self.request_data(api_name, api_path, req_param)

    def delete_blocking_function(self,
                                 path: str,
                                 recursive: Optional[bool] = None,
                                 search_taskid: Optional[str] = None) -> dict[str, object] | str:
        api_name = 'SYNO.FileStation.Delete'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'delete'}

        if type(path) is list:
            new_path = []
            [new_path.append('"' + x + '"') for x in path]
            path = new_path
            path = '[' + ','.join(path) + ']'
            req_param['path'] = path
        elif path is not None:
            req_param['path'] = path
        else:
            return 'Enter a valid path'

        for key, val in locals().items():
            if key not in ['self', 'api_name', 'info', 'api_path', 'req_param', 'path', 'new_path']:
                if val is not None:
                    req_param[str(key)] = val

        'This function will stop your script until done! Do not interrupt '

        return self.request_data(api_name, api_path, req_param)

    def start_extract_task(self,
                           file_path: str,
                           dest_folder_path: str,
                           overwrite: Optional[bool] = None,
                           keep_dir: Optional[bool] = None,
                           create_subfolder: Optional[bool] = None,
                           codepage: Optional[str] = None,
                           password: Optional[str] = None,
                           item_id: Optional[str] = None
                           ) -> dict[str, object] | str:
        api_name = 'SYNO.FileStation.Extract'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'start', 'file_path': file_path,
                     'dest_folder_path': dest_folder_path}

        for key, val in locals().items():
            if key not in ['self', 'api_name', 'info', 'api_path', 'req_param']:
                if val is not None:
                    req_param[str(key)] = val

        if file_path is None:
            return 'Enter a valid file_path'

        if dest_folder_path is None:
            return 'Enter a valid dest_folder_path'

        self._extract_taskid = self.request_data(api_name, api_path, req_param)['data']['taskid']
        self._extract_taskid_list.append(self._extract_taskid)

        message = ('You can now check the status of request with '
                   'get_extract_status() , your id is: '
                   + self._extract_taskid)
        if self.interactive_output:
            output = message
        else:
            output = {"message": message, "taskid": self._extract_taskid}

        return output

    def get_extract_status(self, taskid: str) -> dict[str, object] | str:
        api_name = 'SYNO.FileStation.Extract'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'status'}

        if taskid is None:
            return 'Enter a valid taskid, choose between ' + str(self._extract_taskid_list)
        else:
            req_param['taskid'] = taskid

        return self.request_data(api_name, api_path, req_param)

    def stop_extract_task(self, taskid: str) -> dict[str, object] | str:
        api_name = 'SYNO.FileStation.Extract'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'stop'}

        if taskid is None:
            return 'Enter a valid taskid, choose between ' + str(self._extract_taskid_list)
        else:
            req_param['taskid'] = taskid

        self._extract_taskid_list.remove(taskid)

        return self.request_data(api_name, api_path, req_param)

    def get_file_list_of_archive(self,
                                 file_path: str,
                                 offset: Optional[int] = None,
                                 limit: Optional[int] = None,
                                 sort_by: Optional[str] = None,
                                 sort_direction: Optional[str] = None,
                                 codepage: Optional[str] = None,
                                 password: Optional[str] = None,
                                 item_id: Optional[str] = None
                                 ) -> dict[str, object] | str:
        api_name = 'SYNO.FileStation.Extract'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'list'}

        for key, val in locals().items():
            if key not in ['self', 'api_name', 'info', 'api_path', 'req_param']:
                if val is not None:
                    req_param[str(key)] = val

        if file_path is None:
            return 'Enter a valid file_path'

        return self.request_data(api_name, api_path, req_param)

    def start_file_compression(self,
                               path: str | list[str],
                               dest_file_path: str,
                               level: Optional[int] = None,
                               mode: Optional[str] = None,
                               compress_format: Optional[str] = None,
                               password: Optional[str] = None
                               ) -> dict[str, object] | str | tuple[str]:
        api_name = 'SYNO.FileStation.Compress'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'start'}

        if type(path) is list:
            new_path = []
            [new_path.append('"' + x + '"') for x in path]
            path = new_path
            path = '[' + ','.join(path) + ']'
            req_param['path'] = path
        elif path is not None:
            req_param['path'] = path
        else:
            return 'Enter a valid path'

        for key, val in locals().items():
            if key not in ['self', 'api_name', 'info', 'compress_format', '_password', '_api_path',
                           'req_param', 'path', 'new_path']:
                if val is not None:
                    req_param[str(key)] = val

        if dest_file_path is None:
            return 'Enter a valid dest_file_path'

        if compress_format is not None:
            req_param['format'] = compress_format

        if password is not None:
            req_param['_password'] = password

        self._compress_taskid = self.request_data(api_name, api_path, req_param)['data']['taskid']

        message = ('You can now check the status of request with '
                   'get_compress_status() , your id is: '
                   + self._compress_taskid)
        if self.interactive_output:
            output = message
        else:
            output = {"message": message, "taskid": self._compress_taskid}

        return output

    def get_compress_status(self, taskid: str) -> dict[str, object] | str:
        api_name = 'SYNO.FileStation.Compress'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'status'}

        if taskid is None:
            return 'Enter a valid taskid'
        else:
            req_param['taskid'] = taskid

        return self.request_data(api_name, api_path, req_param)

    def stop_compress_task(self, taskid: str) -> dict[str, object] | str:
        api_name = 'SYNO.FileStation.Compress'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'stop'}

        if taskid is None:
            return 'Enter a valid taskid'
        else:
            req_param['taskid'] = taskid

        return self.request_data(api_name, api_path, req_param)

    def get_list_of_all_background_task(self,
                                        offset: Optional[int] = None,
                                        limit: Optional[int] = None,
                                        sort_by: Optional[str] = None,
                                        sort_direction: Optional[str] = None,
                                        api_filter: Optional[str] = None
                                        ) -> dict[str, object] | str:
        api_name = 'SYNO.FileStation.BackgroundTask'
        info = self.file_station_list[api_name]
        api_path = info['path']
        req_param = {'version': info['maxVersion'], 'method': 'list'}

        for key, val in locals().items():
            if key not in ['self', 'api_name', 'info', 'api_path', 'req_param']:
                if val is not None:
                    req_param[str(key)] = val

        if type(api_filter) is list:
            new_path = []
            [new_path.append('"' + x + '"') for x in api_filter]
            api_filter = new_path
            api_filter = '[' + ','.join(api_filter) + ']'
            req_param['api_filter'] = api_filter

        return self.request_data(api_name, api_path, req_param)

    def get_file(self,
                 path: str,
                 mode: str,
                 dest_path: str = ".",
                 chunk_size: int = 8192,
                 verify: bool = False
                 ) -> Optional[str]:

        api_name = 'SYNO.FileStation.Download'
        info = self.file_station_list[api_name]
        api_path = info['path']

        if path is None:
            return 'Enter a valid path'

        session = requests.session()

        url = ('%s%s' % (self.base_url, api_path)) + '?api=%s&version=%s&method=download&path=%s&mode=%s&_sid=%s' % (
            api_name, info['maxVersion'], parse.quote_plus(path), mode, self._sid)

        if mode is None:
            return 'Enter a valid mode (open / download)'

        if mode == r'open':
            with session.get(url, stream=True, verify=verify) as r:
                r.raise_for_status()
                for chunk in r.iter_content(chunk_size=chunk_size):
                    if chunk:  # filter out keep-alive new chunks
                        sys.stdout.buffer.write(chunk)

        if mode == r'download':
            with session.get(url, stream=True, verify=verify) as r:
                r.raise_for_status()
                if not os.path.isdir(dest_path):
                    os.makedirs(dest_path)
                with open(dest_path + "/" + os.path.basename(path), 'wb') as f:
                    for chunk in r.iter_content(chunk_size=chunk_size):
                        if chunk:  # filter out keep-alive new chunks
                            f.write(chunk)

# TODO SYNO.FileStation.Thumb to be done



"""directory_server.py works with base_api_core to provide AD capabilities."""
import json
from typing import List, Optional, Any
from . import base_api_core


class DirectoryServer(base_api_core.Core):
    """The directory server API.

    Not all items within this class use the Active Directory API.  Some use the Synology Entry API which proxies
    the request. Some are related to managing users in ways that are useful in the Directory Server context.  For
    example, sending a user password reset email, or updating the user information. This api works slightly
    differently than other similar APIs. There are multi-leveled calls where Synology makes requests on behalf of
    the original request and relays information back. Additionally, the query-string api item is not used often in
    this class as API is defined within the actual request.

    The APIs in this class are tested working against the following scenarios:
    - Get Active Directory information
    - List objects within a Base DN on the Active Directory Server
    - Create an AD user
    - Set the user's AD password
    - Send a password reset email to any Synology user
    - Create a new AD group
    - Add an AD user to an AD group
    - Check if an AD object within your Directory Server
    - Update user information within the Directory Server
    - Update Synology's awareness of the current state of the Domain
    - Get the status of a running task such as the Domain status update
    - Delete a list of items from the Directory Server
    - Delete a single item from the Directory Server
    - Perform an entry request to complete a Deletion
    """

    def __init__(self,
                    ip_address: str,
                    port: str,
                    username: str,
                    password: str,
                    secure: bool = False,
                    cert_verify: bool = False,
                    dsm_version: int = 7,
                    debug: bool = True,
                    otp_code: Optional[str] = None
                ) -> None:
        super(DirectoryServer, self).__init__(ip_address, port,
                                              username, password, secure, cert_verify, dsm_version, debug, otp_code)

    def get_directory_info(self) -> dict[str, object] | str:
        """
        Gets directory info.

        Returns
        -------
        Information about your domain. Example below.
        {
            "data": {
                "data": {
                "domainBasicInfo": {
                    "realm": "MY.DOMAIN.COM",
                    "workgroup": "NETBIOSNAME"
                },
                "domainControllers": [
                    {
                    "cn": "AD",
                    "dn": "CN=AD,OU=Domain Controllers,DC=MY,DC=DOMAIN,DC=COM",
                    "dnshostname": "AD.MY.DOMAIN.COM",
                    "roles": [
                        "pdc",
                        "rid",
                        "schema",
                        "naming",
                        "infrastructure"
                    ]
                    }
                ]
                },
                "status": "running"
            },
            "success": true
        }
        """
        api_name = 'SYNO.ActiveDirectory.Info'
        info = {'maxVersion': 3, 'minVersion': 1,
                'path': 'entry.cgi', 'requestFormat': 'JSON'}
        api_path = info['path']
        req_param = {'api': api_name, 'method': 'get',
                     'version': info['maxVersion']}
        return self.request_data(api_name, api_path, req_param)

    def list_directory_objects(self,
                                basedn: str,
                                offset: int = 0,
                                limit: int = 40,
                                objectCategory: list[str] = [ "person", "group", "organizationalUnit", "computer", "container", "builtinDomain"]
                            ) -> dict[str, object] | str:
        """
        lists directory objects.

        Parameters
        ----------
        basedn : str
            The Base DN for the search. eg. "CN=Users,CN=MY,CN=DOMAIN,CN=COM" or CN=MY,CN=DOMAIN,CN=COM

        offset : Optional, int
            When searching large data, you may wish to start at a certain number, e.g. for 10 at a time one
            would set the limit to 10 and the offset by multiples of 10 for each request.
            Default: 0
        limit : Optional, int
            The numeric the number of maximum objects to return.
            Default: 40
        objectCategory : Optional, str([])
            The categories of items to search.  e.g. ["organizationalUnit","container","builtinDomain"] for a list of
            base server containers, and ["person","group","organizationalUnit","computer"] for a list of contained objects.
            Default: ["person","group","organizationalUnit","computer","container","builtinDomain"]

        Returns
        -------
        dictionary
            The result of this method is a dictionary object with a 'data' dictionary and a 'success' dictionary.
            The first level is the success to the AD server.  The second Data level is the status of the actual request.
            Since this is a compound request, the data contains an object with it's own request and results contained
            within. The object will explain any issues with the request.  The data structure is as follows:
            {
                "data": {
                    "data": [
                    {
                        "accountExpiryTime": 910692730085,
                        "deletable": true,
                        "description": "This is a description of a user person",
                        "disabled": false,
                        "displayName": "John Doe",
                        "dn": "CN=jdoe,CN=Users,DC=MY,DC=DOMAIN,DC=COM",
                        "locked": false,
                        "mail": "jdoe@MY.EMAIL.COM",
                        "movable": true,
                        "name": "john",
                        "objectCategory": "person",
                        "passwordExpired": true,
                        "physicalDeliveryOfficeName": "official office of officers",
                        "primaryGroupToken": 0,
                        "renamable": true,
                        "sAMAccountName": "jdoe",
                        "showInAdvancedViewOnly": false,
                        "telephoneNumber": "123-444-5677"
                    },
                    ],
                    "total": 99999
                },
                "success": true
            }
        """
        action = '"enum"'
        scope = '"one"'
        api = 'SYNO.ActiveDirectory.Directory'
        method = '"list"'
        info = {'maxVersion': 3, 'minVersion': 1,
                'path': 'entry.cgi', 'requestFormat': 'JSON'}
        api_path = info['path']
        req_param = {'api': api, 'action': action, 'basedn': '"'+basedn+'"', 'limit': limit, "method": method,
                     'objectCategory': json.dumps(objectCategory), 'offset': offset, 'scope': scope,  'version': info['maxVersion']}
        print(req_param)
        return self.request_data("", api_path, req_param, 'post')

    def create_new_user(
            self,
            logon_name: str,
            email: str,
            password: str,
            located_dn: str,
            description: str = '',
            account_is_disabled: str = 'false',
            cannot_change_password: str = 'false',
            change_password_next_logon: str = 'null',
            password_never_expire: str = 'true'
    ) -> List[str]:
        """Create a new user.

        Please note that synchronization with Synology is a separate step.  The user can be created in AD, but not able to log on until the next synchronization occurs.

        Parameters
        ----------
        logon_name : str
            The desired username.  "jdoe"
        email: str
            The desired email
        password : str
            The plain-text password for the new user.  "Password123"
        located_dn : str
            The DN for the user.  "CN=Users,CN=MY,CN=DOMAIN,CN=COM"
        description : str, optional
            A description for the user
        account_is_disabled : str
            Set to 'true' if the account should be disabled (default is false)
        cannot_change_password : str, optional
            Set to 'true' if the user cannot change the password (default is false)
        change_password_next_logon : str, optional
            Set to 'true' if the user must change password on next logon (default is false)
        cannot_change_password : str, optional
            Set to 'true' if the user cannot change the password (default is false)
        password_never_expire: str
            Pwd Never Expire

        Returns
        -------
        dictionary
            The result of this method is a dictionary object with a 'data' dictionary and a 'success' dictionary. The
            data dictionary contains an 'error', or it contains a 'dn' and a 'name'. here is an example of a successful
            result.
            {'data': {'dn': 'CN=jdoe,CN=Users,DC=MY,DC=DOMAIN,DC=COM', 'name': 'NETBIOSNAME\\ababab'}, 'success': True}
        """

        api_name = "SYNO.ActiveDirectory.User"
        info = {'maxVersion': 1, 'minVersion': 1,
                'path': 'entry.cgi', 'requestFormat': 'JSON'}
        api_path = info['path']
        req_param = {'api': api_name, 'version': info['maxVersion'], 'method': 'create', 'logon_name': logon_name, 'email': email,
                     'located_dn': located_dn, 'password': password,  'description': description, 'account_is_disabled': account_is_disabled,
                     'cannot_change_password': cannot_change_password, 'change_password_next_logon': change_password_next_logon,
                     'password_never_expire': password_never_expire}
        return self.request_data(api_name, api_path, req_param)

    def reset_password(self,
                       username: str,
                       ) -> List[str]:
        """
        Send a password reset email.

        This will trigger the password reset email from
        Control Panel>Notification>Rules>System>Reset password for your account to be sent
        to the user. In order to use this,
        Control Panel>User & Group>Advanced>"Allow non-administrator users to reset forgotten passwords via email"
        must be enabled.

        Parameters
        ----------
        username : str
            The username to reset.  E.g. "My Group"

        Returns
        -------
        dictionary
           The return object can be checked for the "success" to be a true or false.
           True indicates a successful operation.

            {"data": {"msg": 3}, "success": true}

        """

        api_name = 'SYNO.Auth.ForgotPwd'
        newApi = {'maxVersion': 1, 'minVersion': 1,
                  'path': 'entry.cgi', 'requestFormat': 'JSON'}
        info = newApi
        api_path = info['path']
        req_param = {'method': 'send', 'user': '"' +
                     username+'"', 'version': newApi['maxVersion']}
        return self.request_data(api_name, api_path, req_param)

    def change_user_password(self, user_dn: str, password: str) -> dict[str, object] | str:
        """
        Change the user's password.  This is a compound dual-level request where the synology API proxies your
        request to the Directory Server.

        Parameters
        ----------
        user_dn: str
            The user DN to be modified. eg. "CN=jdoe,CN=Users,DC=MY,DC=DOMAIN,DC=COM"
        password: str
            The new password to be set. e.g. "Password123"

        Returns
        -------
            The result of this method is a dictionary object with a 'data' dictionary and a 'success' dictionary.
            The first level is the success to the AD server.  The second Data level is the status of the actual request.
            Since this is a compound request, the data contains an object with it's own request and results contained
            within. The object will explain any issues with the request.  The data structure is as follows:
            {
                 "data": {
                    "has_fail": false,
                    "result": [
                    {
                        "api": "SYNO.ActiveDirectory.User",
                        "data": [
                        {
                            "code": 0,
                            "msg": "update record successfully"
                        }
                        ],
                        "method": "set",
                        "success": true,
                        "version": 2
                    }
                    ]
                },
                "success": true
            }

        """
        api_name = "SYNO.Entry.Request"
        info = {'maxVersion': 1, 'minVersion': 1,
                'path': 'entry.cgi', 'requestFormat': 'JSON'}
        compound = '[{"api":"SYNO.ActiveDirectory.User","method":"set","version":2,"userList":[{"dn":"' + \
            user_dn+'","enbl_change_password":true,"password":"'+password+'"}]}]'
        api_path = info['path']
        req_param = {"api": api_name, 'method': 'request', 'compound': compound,
                     'mode': 'sequential', 'stop_when_error': "true", 'version': info['maxVersion']}
        return self.request_data(api_name, api_path, req_param)

    def create_new_group(
            self,
            name: str,
            located_dn: str,
            email: Optional[str] = '',
            description: Optional[str] = '',
            type: Optional[str] = 'security',
            scope: Optional[str] = 'global'
    ) -> List[str]:
        """
        Create a new AD group.

        Parameters
        ----------
        name : str
            The name of the group.  E.g. "My Group"
        located_dn : str
            The DN to place the group in.  eg. "CN=Groups,DC=MY,DC=DOMAIN,DC=COM"
        email : str, Optional
            The email address used to reference this group.
            Default: ""
        description : str, Optional
            A description of the AD Group.
            Default: Empty
        type : str, Optional
            Example Options: security, distribution

            (definitions from
            https://docs.microsoft.com/en-us/microsoft-365/admin/create-groups/compare-groups?view=o365-worldwide
            )
                - distribution (Distribution groups) are used for sending email
                notifications to a group of people.
                - security - Security groups are used for granting access to resources
                such as SharePoint sites.

            Default: security
        scope : str, Optional
            Example Options : local, global, universal
            (Definitions from
            https://www.netwrix.com/active_directory_group_management.html )
                - local (Domain Local Groups) should be used to manage permissions to
                resources because this group can be applied everywhere in the domain.
                A domain local group can include members of any type in the domain and
                members from trusted domains. For example, suppose you need access
                management for a collection of folders on one or more servers that
                contain information for managers. The group you create for that purpose
                should be a domain local group (ex. “DL_Managers_Modify”).
                - global (Global Groups) are used primarily to define collections of
                domain objects (users, other global groups and computers) based on
                business roles, which means that they mostly serve as role groups.
                Role-based groups of users (such as “HR” or “Marketing”) and role-based
                groups of computers (such as a “Marketing Workstations”) areusually
                global groups.
                - universal (Universal Groups) in Active Directory are useful in
                multi-domain forests. They enable you to define roles or manage
                resources that span more than one domain. Each universal group is
                stored in the domain of where it was created, but its group membership
                is stored in the Global Catalog and replicated forest-wide. Don’t use
                universal groups if you have only one domain.

            Default: global

        Returns
        -------
            A success object, and data object containing the new dn and the netbios name of the group.
            {'data': {'dn': 'CN=My Group,CN=Groups,DC=MY,DC=DOMAIN,DC=COM', 'name': 'NETBIOSNAME\\My Group'}, 'success': True}
        """
        api_name = 'SYNO.ActiveDirectory.Group'
        info = {'maxVersion': 1, 'minVersion': 1,
                'path': 'entry.cgi', 'requestFormat': 'JSON'}
        api_path = info['path']
        req_param = {'api': api_name, 'method': 'create', 'name': name, 'located_dn': located_dn,
                     'description': description, 'type': type, 'scope': scope, 'email': email, 'version': info['maxVersion']}
        return self.request_data(api_name, api_path, req_param)

    def add_user_to_group(self, userDn: str, groupDn: str) -> dict[str, object] | str:
        """
        Adds a user as a member of a group.


        Parameters
        ----------
        userDn : str
            The fully qualified dn to add.  eg. "CN=jdoe,CN=Users,CN=MY,CN=DOMAIN,CN=COM"

        groupDn : str
            the fully qualified dn of the group to which the user is to be added.
            e.g. "CN=My Group,CN=Groups,CN=MY,CN=DOMAIN,CN=COM"

        Returns
        -------
        dictionary
            The result of this method is a dictionary object with a 'data' dictionary and a 'success' dictionary.
            The first level is the success to the AD server.  The second Data level is the status of the actual request.
            Since this is a compound request, the data contains an object with it's own request and results contained
            within. The object will explain any issues with the request.  The data structure is as follows:
            {
                "data": {
                    "has_fail": false,
                    "result": [
                    {
                        "api": "SYNO.ActiveDirectory.Group.Member",
                        "data": {
                        "members": [
                            "CN=jdoe,CN=Users,CN=MY,CN=DOMAIN,CN=COM"
                        ]
                        },
                        "method": "add",
                        "success": true,
                        "version": 1
                    }
                    ]
                },
                "success": true
            }


        """

        api_name = 'SYNO.Entry.Request'
        compound = '[{"api":"SYNO.ActiveDirectory.Group.Member","method":"add","version":"1","dn":"' + \
            groupDn+'","members":["'+userDn+'"]}]'
        method = 'request'
        mode = "sequential"
        stop_when_error = True
        newApi = {'maxVersion': 1, 'minVersion': 1,
                  'path': 'entry.cgi', 'requestFormat': 'JSON'}
        info = newApi
        api_path = info['path']
        req_param = {'api': api_name, 'compound': compound, 'method': method, 'mode': mode,
                     'stop_when_error': stop_when_error, 'version': newApi['maxVersion']}
        return self.request_data(api_name, api_path, req_param)

    def does_dn_exist(self, groupName: str) -> dict[str, object] | str:
        """Checks if a container exists. This can be used to verifiy the username or group name is unique.  This will
        not check the container, only if a similarly named container already exists.

        Parameters
        ----------
        groupName : str
            The user, or group's name. e.g.  "jdoe" or "My Cool Group"
            Fully Qualified Domain Name such as "CN=My Cool Group,CN=Groups,DC=MY,DC=DOMAIN,DC=COM" are not successful
            Improper case such as "my cool group" instead of "My Cool Group" are successful

        Returns
        -------
        boolean
            True if the group exists.  False if the group does not exist
        """

        api_name = 'SYNO.ActiveDirectory.Group'
        info = {'maxVersion': 1, 'minVersion': 1,
                'path': 'entry.cgi', 'requestFormat': 'JSON'}
        api_path = info['path']
        req_param = {'version': info['maxVersion'],
                     'method': 'conflict', 'name': groupName}
        return self.request_data(api_name, api_path, req_param)['data']['isConflict']

    def modify_user_info(self,
        user_dn: str = None,
        firstName: str = None,
        lastName: str = None,
        displayName: str = None,
        description: str = None,
        initials: str = None,
        physicalDeliveryOfficeName: str = None,
        telephoneNumber: str = None,
        web: str = None
    ) -> Any:
        """
        Performs modification to user information within the Active Directory.

        Parameters
        ----------
        user_dn: str
            The user DN to be modified. eg. "CN=jdoe,CN=Users,DC=MY,DC=DOMAIN,DC=COM"
        firstName: Optional, str
            The First name of the user. e.g. "John"
        lastName: Optional, str
            The Last Name of the user. e.g. "Doe"
        displayName: Optional, str
            The Display name of the user. e.g. "John Doe"
        description: Optional, str
            The Descrition of the user. e.g. "The guy who just came in"
        initials: Optional, str
            The Initials of the user.  e.g. "JD"
        physicalDeliveryOfficeName: Optional, str
            The office location in the user's place of business
        telephoneNumber: Optional, str
            The user's telephone number.
        web: Optional, str
            The user's website or location on the web where information can be obtained.

        Returns
        -------
            The result of this method is a dictionary object with a 'data' dictionary and a 'success' dictionary.
            The first level is the success to the AD server.  The second Data level is the status of the actual request.
            Since this is a compound request, the data contains an object with it's own request and results contained
            within. The object will explain any issues with the request.  The data structure is as follows:
            {
                "data": {
                    "has_fail": true,
                    "result": [
                    {
                        "api": "SYNO.ActiveDirectory.User",
                        "error": {
                        "code": 10104,
                        "errors": [
                            {
                            "code": 10237,
                            "msg": "ldb updaterecords: modify"
                            }
                        ]
                        },
                        "method": "set",
                        "success": false,
                        "version": 2
                    }
                    ]
                },
                "success": true
            }
        """
        class Person:
            firstName
            lastName
            displayName
            description
            initials
            physicalDeliveryOfficeName
            telephoneNumber
            web
            user_dn
        userObject = Person()
        userObject.dn = user_dn
        if firstName is not None:
            userObject.firstName = firstName
        if lastName is not None:
            userObject.lastName = lastName
        if displayName is not None:
            userObject.displayName = displayName
        #if description is not None:
        #    userObject.description = description
        if initials is not None:
            userObject.initials = initials
        if physicalDeliveryOfficeName is not None:
            userObject.physicalDeliveryOfficeName = physicalDeliveryOfficeName
        if telephoneNumber is not None:
            userObject.telephoneNumber = telephoneNumber
        if web is not None:
            userObject.web = web

        theJsonObject = userObject.__dict__
        val = self.setEntryRequest("SYNO.ActiveDirectory.User", "set", "userList", theJsonObject)

        return val

    def setEntryRequest(self, modificationAPI: str, method: str, nameOfObject: str, jsonObject: Any) -> dict[str, object] | str:
        """
        Performs modification to an object within the Active Directory.

        Parameters
        ----------
        modificationAPI: str
        method: str
        nameOfObject: str
            The user DN to be modified. eg. "CN=jdoe,CN=Users,DC=MY,DC=DOMAIN,DC=COM"
        jsonObject: str: o
            the json Object to be added, eg, a user object where the

        Returns
        -------
            The result of this method is a dictionary object with a 'data' dictionary and a 'success' dictionary.
            The first level is the success to the AD server.  The second Data level is the status of the actual request.
            Since this is a compound request, the data contains an object with it's own request and results contained
            within. The object will explain any issues with the request.  The data structure is as follows:
            {
                "data": {
                    "has_fail": true,
                    "result": [
                    {
                        "api": "SYNO.ActiveDirectory.User",
                        "error": {
                        "code": 10104,
                        "errors": [
                            {
                            "code": 10237,
                            "msg": "ldb updaterecords: modify"
                            }
                        ]
                        },
                        "method": "set",
                        "success": false,
                        "version": 2
                    }
                    ]
                },
                "success": true
            }
        """
        compound = [{"api":modificationAPI,"method":  method,"version":2,nameOfObject:[jsonObject]}]
        api_name = "SYNO.Entry.Request"
        info = {'maxVersion': 1, 'minVersion': 1,
                'path': 'entry.cgi', 'requestFormat': 'JSON'}
        api_path = info['path']
        req_param = {"api": api_name, 'method': 'request', 'compound': json.dumps(compound), 'mode': '"sequential"', 'stop_when_error': True, 'version': info['maxVersion']}
        print (json.dumps(req_param))
        return self.request_data(api_name, api_path, req_param,"post")

    def update_domain_records(self) -> dict[str, object] | str:
        """
        Updates the Synology users and groups database with information from Directory Server.

        This is a long-running and asynchronous task.  You are given back a task_id, and you can
        use that task_id to check the status with the get_task_status(task_id) method.

        Returns
        -------
        dictionary
            The 'data' object contains the 'task_id' used to track with the getTaskStatus() method.
            The 'success' object will be true if the operation was successful. or false if failed.

            {"data": {"task_id": "@administrators/DomainUpdate6146195136397F2"}, "success": true}

        Note
        ----
        Typical utilization of Update Domain requires starting the update job and waiting for
        completion. Waiting involves using the getTaskStatus and can be accomplished via a busy-wait
        method such as the following:

            updateResponse=directory.updateDomain()
            status=directory.getTaskStatus(updateResponse['data']['task_id'])
            while status['data']['status'] == 'updating' :
                status=directory.getTaskStatus(updateResponse['data']['task_id'])
        """
        api_name = 'SYNO.Core.Directory.Domain'
        info = self.core_list[api_name]
        api_path = info['path']
        req_param = {'domain_name': '"@all"',
                     'method': 'update_start', 'version': info['minVersion']}
        return self.request_data(api_name, api_path, req_param)

    def get_task_status(self, task_id: str) -> dict[str, object] | str:
        """
        Gets the current status of a task running on the Directory Domain object.

        This is used to ensure the task is completed.  For example, the primary utilization of this is
        when updating Synology's internal Domain user and group list.  Until this method reports
        finish, the job is not completed, and it is not safe to operate under the assumption that users
        have been synchronized.

        Parameters
        ----------
        task_id : str
            The task ID to be tracked for status.

        Returns
        -------
        dictionary
            The 'data' object contains the 'status' used to determine the current status. 'status'
            will be 'updating' or 'finish' if the job was started.
            The 'success' object will be true if the operation was successful. or false if failed.

            {'data': {'status': 'updating'}, 'success': True}
        """

        api_name = 'SYNO.Core.Directory.Domain'
        info = self.core_list[api_name]
        api_path = info['path']
        req_param = {'method': 'update_status',
                     'task_id': task_id, 'version': info['minVersion']}
        return self.request_data(api_name, api_path, req_param)

    def deleteItems(self, dnList: list[str]) -> dict[str, object] | str:
        """
        Deletes an array of DNs from AD.

        Parameters
        ----------
        dnList : str([])
            The fully qualified DN to be removed from the directory server.
            eg. ["CN=jdoe,CN=Users,CN=MY,CN=DOMAIN,CN=COM","CN=My Group,CN=Groups,CN=MY,CN=DOMAIN,CN=COM"]

        Returns
        -------
        The result of this method is a dictionary object with a 'data' dictionary and a 'success' dictionary.
        The first level is the success to the AD server.  The second Data level is the status of the actual request.
        Since this is a compound request, the data contains an object with it's own request and results contained
        within. The object will explain any issues with the request.  The data structure is as follows:

            {
                "data": {
                    "has_fail": false,
                    "result": [
                    {
                        "api": "SYNO.ActiveDirectory.Polling",
                        "data": {
                        "data": [
                            {
                            "dn": "CN=My Group,CN=Groups,CN=MY,CN=DOMAIN,CN=COM",
                            "status": {}
                            }
                        ],
                        "finished": true,
                        "total": 1
                        },
                        "method": "get",
                        "success": true,
                        "version": 1
                    }
                    ]
                },
                "success": true
            }
        """
        api_name = 'SYNO.ActiveDirectory.Directory'
        info = {'maxVersion': 2, 'minVersion': 1,
                'path': 'entry.cgi', 'requestFormat': 'JSON'}
        api_path = info['path']
        req_param = {'api': api_name, 'method': 'delete',
                     'dnList': json.dumps(dnList), 'version': 2}
        task_id = self.request_data(api_name, api_path, req_param)[
            'data']['task_id']
        returnValue = self.entryRequest(task_id)
        notFinished = True
        while 'data' in returnValue['data']['result'][0] and notFinished:
            notFinished = False
            for resultItem in returnValue['data']['result']:
                if not resultItem['data']['finished']:
                    notFinished = True
            if not notFinished:
                break
            returnValue = self.entryRequest(task_id)

        return returnValue

    def delete_item(self, dn: str) -> dict[str, object] | str:
        """
        Deletes a DN from AD.

        Parameters
        ----------
        dn : str
            The fully qualified DN to be removed from the directory server.
            eg. "CN=jdoe,CN=Users,CN=MY,CN=DOMAIN,CN=COM" or
            "CN=My Group,CN=Groups,CN=MY,CN=DOMAIN,CN=COM"

        Returns
        -------
        The result of this method is a dictionary object with a 'data' dictionary and a 'success' dictionary.
        The first level is the success to the AD server.  The second Data level is the status of the actual request.
        Since this is a compound request, the data contains an object with it's own request and results contained
        within. The object will explain any issues with the request.  The data structure is as follows:

            {
                "data": {
                    "has_fail": false,
                    "result": [
                    {
                        "api": "SYNO.ActiveDirectory.Polling",
                        "data": {
                        "data": [
                            {
                            "dn": "CN=My Group,CN=Groups,CN=MY,CN=DOMAIN,CN=COM",
                            "status": {}
                            }
                        ],
                        "finished": true,
                        "total": 1
                        },
                        "method": "get",
                        "success": true,
                        "version": 1
                    }
                    ]
                },
                "success": true
            }
        """
        items = []
        items.append(dn)
        return self.deleteItems(items)

    def entryRequest(self, task_id: str) -> Any:
        """
        Some requests require an entry.

        Delete for example requires an entry.  If an entry is required, the task will not
        complete without an Entry Request.

        Parameters
        ----------
        task_id: str
            The ID of the task to be checked. This is provided when making a request. An example Task ID may look like this
            "@administrators/Synoads_SYNO.ActiveDirectory.Directory_delete6145EA17C4F03DA9"
        """
        api_name = 'SYNO.Entry.Request'
        info = {'maxVersion': 1, 'minVersion': 1,
                'path': 'entry.cgi', 'requestFormat': 'JSON'}
        compound = '[{"api":"SYNO.ActiveDirectory.Polling","method":"get","version":1,"task_id":"'+task_id+'"}]'
        api_path = info['path']
        req_param = {'api': api_name, 'method': 'request',  'compound': compound,
                     'mode': 'parallel', 'version': info['maxVersion']}
        return self.request_data(api_name, api_path, req_param)



from io import BytesIO
from typing import Optional

from . import base_api_core

import os
import requests
import json


class Certificate(base_api_core.Core):
    def __init__(self,
                 ip_address: str,
                 port: str,
                 username: str,
                 password: str,
                 secure: bool = False,
                 cert_verify: bool = False,
                 dsm_version: int = 7,
                 debug: bool = True,
                 otp_code: Optional[str] = None
                 ) -> None:
        super(Certificate, self).__init__(ip_address, port, username, password, secure, cert_verify, dsm_version, debug,
                                          otp_code)
        self._debug: bool = debug
        return

    def _base_certificate_methods(self,
                                  method: str,
                                  cert_id: Optional[str] = None,
                                  ids: Optional[str | list[str]] = None
                                  ) -> str | dict[str, object]:
        available_method = ['list', 'set', 'delete']
        if method not in available_method:
            # print error here
            return f"Method {method} no supported."

        api_name = 'SYNO.Core.Certificate.CRT'
        info = self.gen_list[api_name]
        api_path = info['path']

        req_param = {'version': '1',
                     'method': method}

        if 'set' == method and cert_id:
            req_param.update(
                {'as_default': 'true',
                 'desc': '\"\"',
                 'id': f"\"{cert_id}\""
                 })
        elif 'delete' == method and ids:
            ids = json.dumps(ids)
            req_param.update({"ids": ids})

        return self.request_data(api_name, api_path, req_param)

    def list_cert(self) -> dict[str, object]:
        return self._base_certificate_methods('list')

    def set_default_cert(self, cert_id: str) -> dict[str, object]:
        return self._base_certificate_methods('set', cert_id)

    def delete_certificate(self, ids: str | list[str]) -> dict[str, object]:
        if isinstance(ids, str):
            ids = [ids]
        return self._base_certificate_methods('delete', ids=ids)

    def upload_cert(self,
                    serv_key: str = "server.key",
                    ser_cert: str = "server.crt",
                    ca_cert: Optional[str] = None,
                    set_as_default: bool = True,
                    cert_id: Optional[str] = None,
                    desc: Optional[str] = None
                    ) -> tuple[int, dict[str, object]]:
        api_name = 'SYNO.Core.Certificate'
        info = self.session.app_api_list[api_name]
        api_path = info['path']
        serv_key = os.path.abspath(serv_key)
        ser_cert = os.path.abspath(ser_cert)
        # ca_cert is optional argument for upload cert
        ca_cert = os.path.abspath(ca_cert) if ca_cert else None

        session = requests.session()

        url = ('%s%s' % (self.base_url, api_path)) + '?api=%s&version=%s&method=import&_sid=%s' % (
            api_name, info['minVersion'], self._sid)

        if cert_id:
            print("update exist cert: " + cert_id)
        data_payload = {'id': cert_id or '', 'desc': desc or '', 'as_default': 'true' if set_as_default else 'false'}

        with open(serv_key, 'rb') as payload_serv_key, open(ser_cert, 'rb') as payload_ser_cert:
            files = {'key': (serv_key, payload_serv_key, 'application/x-iwork-keynote-sffkey'),
                     'cert': (ser_cert, payload_ser_cert, 'application/pkix-cert')}
            if ca_cert:
                with open(ca_cert, 'rb') as payload_ca_cert:
                    files['inter_cert'] = (ca_cert, payload_ca_cert, 'application/pkix-cert')
                    r = session.post(url, files=files, data=data_payload, verify=self.session.verify_cert_enabled())
            else:
                r = session.post(url, files=files, data=data_payload, verify=self.session.verify_cert_enabled())

        if 200 == r.status_code and r.json()['success']:
            if self._debug is True:
                print('Certificate upload successful.')

        return r.status_code, r.json()

    def export_cert(self, cert_id: str) -> Optional[BytesIO]:
        """Export a certificate from the Synology NAS.

        :param cert_id: The certificate ID to export. This can be found in the list_cert() method.
        :return: A BytesIO object containing the certificate archive.
        """

        api_name = "SYNO.Core.Certificate"
        info = self.session.app_api_list[api_name]
        api_path = info['path']

        session = requests.session()

        url = (
            f"{self.base_url}{api_path}?"
            f"api={api_name}&"
            f"version={info['minVersion']}&"
            f"method=export&"
            f"file=\"archive\"&"
            f"_sid={self._sid}&"
            f"id={cert_id}"
        )

        result = session.get(url, verify=self.session.verify_cert_enabled())

        if result.status_code == 200:
            return BytesIO(result.content)

        return
